Imports System.IO
' The Bistream class is responsible for parsing an MPEG audio bitstream.
Public Class Bitstream

#Region " Public Variables "
    ' Synchronization control constant for the initial synchronization 
    ' to the start of a frame
    Public Shared INITIAL_SYNC As Short = 0
    ' Synchronization control constant for non-initial frame synchronizations
    Public Shared STRICT_SYNC As Short = 1
#End Region

#Region " Private Variables "
    ' The frame buffer that holds the data for the current frame
    Private m_framebuffer(432) As Integer
    ' Number of valid bytes in the frame buffer
    Private m_framesize As Integer
    ' The bytes read from the stream
    Private m_frame_bytes(1731) As Byte
    ' Index into framebuffer where the next bits are retrieved
    Private m_wordpointer As Integer
    ' Number (0-31, from MSB to LSB) of next bit for get_bits()
    Private m_bitindex As Integer
    ' The current specified syncword
    Private m_syncword As Integer
    Private m_single_ch_mode As Boolean
    Private m_bitmask() As Integer = {0, &H1, &H3, &H7, &HF, &H1F, &H3F, &H7F, &HFF, &H1FF, &H3FF, &H7FF, &HFFF, &H1FFF, &H3FFF, &H7FFF, &HFFFF, &H1FFFF}
    ' The InputStream to read from
    Private m_backstream As BackStream
    Private m_header As FrameHeader
    Private m_syncbuf(3) As Byte
#End Region

#Region " Constructor "
    ' Construct a Bitstream that reads data from a given InputStream
    Friend Sub New(ByVal xBackStream As BackStream)
        If xBackStream Is Nothing Then
            Throw New NullReferenceException("BackStream")
        End If
        m_header = New FrameHeader
        m_backstream = xBackStream
        CloseFrame()
    End Sub
#End Region

#Region " Functions "
    ' Reads and parses the next frame from the input source.
    ' returns the Header describing details of the frame read,
    ' or null if the end of the stream has been reached.
    Friend Function ReadFrame() As FrameHeader
        Dim result As FrameHeader = Nothing
        Try
            result = ReadNextFrame()
        Catch ex As BitstreamException
            If ex.ErrorCode <> BitstreamErrors.STREAM_EOF Then
                Throw NewBitstreamException(ex.ErrorCode, ex)
            End If
        End Try
        Return result
    End Function

    Private Function ReadNextFrame() As FrameHeader
        If m_framesize = -1 Then
            m_header.ReadHeader(Me)
        End If

        Return m_header
    End Function

    ' Reads the exact number of bytes from the source input stream into a byte array.
    Private Sub ReadFully(ByVal b() As Byte, ByVal offs As Integer, ByVal len As Integer)
        Try
            While len > 0
                Dim bytesread As Integer = m_backstream.Read(b, offs, len)
                If bytesread = -1 OrElse bytesread = 0 Then
                    While len > 0
                        b(offs) = 0
                        offs += 1
                        len -= 1
                    End While
                    Exit While
                End If

                offs += bytesread
                len -= bytesread
            End While
        Catch ex As IOException
            Throw NewBitstreamException(BitstreamErrors.STREAM_ERROR, ex)
        End Try

    End Sub

    ' Simlar to readFully, but doesn't throw exception when EOF is reached.
    Private Function ReadBytes(ByVal b() As Byte, ByVal offs As Integer, ByVal len As Integer) As Integer
        Dim totalBytesRead As Integer = 0
        Try
            While len > 0
                Dim bytesread As Integer = m_backstream.Read(b, offs, len)
                If bytesread = -1 OrElse bytesread = 0 Then
                    Exit While
                End If
                totalBytesRead += bytesread
                offs += bytesread
                len -= bytesread
            End While
        Catch ex As IOException
            Throw NewBitstreamException(BitstreamErrors.STREAM_ERROR, ex)
        End Try
        Return totalBytesRead

    End Function

    ' Reads the data for the next frame. The frame is not parsed until parse 
    ' frame is called.
    Friend Sub ReadFrameData(ByVal bytesize As Integer)
        ReadFully(m_frame_bytes, 0, bytesize)
        m_framesize = bytesize
        m_wordpointer = -1
        m_bitindex = -1

    End Sub

    ' Get next 32 bits from bitstream. They are stored in the headerstring.
    ' syncmode allows Synchro flag ID. The returned value is False at the end of stream.
    Friend Function SyncHeader(ByVal syncmode As Short) As Integer
        Dim sync As Boolean
        Dim headerstring As Integer
        ' read additinal 2 bytes
        Dim bytesRead As Integer = ReadBytes(m_syncbuf, 0, 3)
        If bytesRead <> 3 Then
            Throw NewBitstreamException(BitstreamErrors.STREAM_EOF, Nothing)
        End If

        headerstring = ((CInt(m_syncbuf(0)) << 16) And &HFF0000) Or ((CInt(m_syncbuf(1)) << 8) And &HFF00) Or ((CInt(m_syncbuf(2)) << 0) And &HFF)

#If THROW_ON_SYNC_LOSS Then
        Dim bytesSkipped As Integer = 0
        Dim lostSyncYet As Boolean = False
#End If

        Do
            headerstring = (headerstring << 8)

            If ReadBytes(m_syncbuf, 3, 1) <> 1 Then
                Throw NewBitstreamException(BitstreamErrors.STREAM_EOF, Nothing)
            End If

            headerstring = headerstring Or (m_syncbuf(3) And &HFF)
            sync = IsSyncMark(headerstring, syncmode, m_syncword)


#If THROW_ON_SYNC_LOSS Then
            If Not sync AndAlso Not lostSyncYet Then
                lostSyncYet = True
                Trace.WriteLine("Lost Sync :(", "Bitstream")
            End If

            If lostSyncYet AndAlso sync Then
                Trace.WriteLine("Found Sync", "Bitstream")
            End If


            bytesSkipped += 1
            If bytesSkipped Mod 2048 = 0 Then
                Trace.WriteLine("Sync still not found", "Bitstream")
            End If
#End If

        Loop While Not sync

        Return headerstring

    End Function

    ' Determines if the next 4 bytes of the stream represent a frame header
    Public Function IsSyncCurrentPosition(ByVal syncmode As Integer) As Boolean
        Dim read As Integer = ReadBytes(m_syncbuf, 0, 4)
        Dim headerstring As Integer = Common.CreateIntegerFromBytes(m_syncbuf)

        Try
            m_backstream.UnRead(read)
        Catch
        End Try

        Dim sync As Boolean = False
        Select Case read
            Case 0
                sync = True

            Case 4
                sync = IsSyncMark(headerstring, syncmode, m_syncword)
        End Select

        Return sync

    End Function

    Public Function IsSyncMark(ByVal headerstring As Integer, ByVal syncmode As Integer, ByVal word As Integer) As Boolean
        Dim sync As Boolean = False
        If syncmode = INITIAL_SYNC Then
            sync = ((headerstring And &HFFE00000) = &HFFE00000)
        Else
            Dim ZW As Boolean = ((headerstring And &HC0) = &HC0)
            sync = ((headerstring And &HFFE00000) = &HFFE00000) AndAlso (ZW = m_single_ch_mode)
        End If

        If sync Then
            ' filter out invalid sample rate
            sync = (Common.URShift(headerstring, 10) And 3) <> 3
        End If

        If sync Then
            ' filter out invalid layer
            sync = (Common.URShift(headerstring, 17) And 3) <> 0
        End If
        If sync Then
            ' filter out invalid version
            sync = (Common.URShift(headerstring, 19) And 3) <> 1
            If Not sync Then
                Console.WriteLine("INVALID VERSION DETECTED")
            End If
        End If
        Return sync

    End Function


    ' Parses the data previously read with read_frame_data()
    Friend Sub ParseFrame()
        ' Convert Bytes read to Integer
        Dim b As Integer = 0
        Dim byteread As Byte() = m_frame_bytes
        Dim bytesize As Integer = m_framesize
        Dim k As Integer = 0
        Dim bx(3) As Byte

        While k < bytesize

            bx(0) = byteread(k)
            If k + 1 < bytesize Then
                bx(1) = byteread((k + 1))
            Else
                bx(1) = 0
            End If
            If k + 2 < bytesize Then
                bx(2) = byteread((k + 2))
            Else
                bx(2) = 0
            End If
            If k + 3 < bytesize Then
                bx(3) = byteread((k + 3))
            Else
                bx(3) = 0
            End If
            m_framebuffer(b) = Common.CreateIntegerFromBytes(bx)
            b += 1
            k = k + 4
        End While

        m_wordpointer = 0
        m_bitindex = 0

    End Sub

    ' Set the word we want to sync the header to in Big-Endian byte order.
    Friend Sub SetSyncword(ByVal syncword0 As Integer)
        m_syncword = syncword0 And &HFFFFFF3F
        m_single_ch_mode = (syncword0 And &HC0) = &HC0
    End Sub

    ' Read bits from buffer into the lower bits of an Integer.
    ' The LSB contains the latest read bit of the stream.
    Public Function GetBits(ByVal number_of_bits As Integer) As Integer
        Dim returnvalue As Integer = 0
        Dim sum As Integer = m_bitindex + number_of_bits

        If m_wordpointer < 0 Then
            m_wordpointer = 0
        End If
        If sum <= 32 Then
            ' all bits contained in *wordpointer
            returnvalue = Common.URShift(m_framebuffer(m_wordpointer), 32 - sum) And m_bitmask(number_of_bits)
            m_bitindex += number_of_bits
            If m_bitindex = 32 Then
                m_bitindex = 0
                m_wordpointer += 1
            End If
            Return returnvalue
        End If

        Dim Right As Integer = m_framebuffer(m_wordpointer) And &HFFFF
        m_wordpointer += 1
        Dim Left As Integer = m_framebuffer(m_wordpointer) And &HFFFF0000
        returnvalue = ((Right << 16) And &HFFFF0000) Or (Common.URShift(Left, 16) And &HFFFF)

        returnvalue = Common.URShift(returnvalue, 48 - sum)
        returnvalue = returnvalue And m_bitmask(number_of_bits)
        m_bitindex = sum - 32
        Return returnvalue
    End Function

    ' Unreads the bytes read from the frame
    Public Sub UnreadFrame()
        If m_wordpointer = -1 AndAlso m_bitindex = -1 AndAlso m_framesize > 0 Then
            Try
                m_backstream.UnRead(m_framesize)
            Catch
                Throw NewBitstreamException(BitstreamErrors.STREAM_ERROR)
            End Try
        End If

    End Sub

    Public Sub CloseFrame()
        m_framesize = -1
        m_wordpointer = -1
        m_bitindex = -1
    End Sub

    Public Sub Close()
        Try
            m_backstream.Close()
        Catch ex As IOException
            Throw NewBitstreamException(BitstreamErrors.STREAM_ERROR, ex)
        End Try
    End Sub

    Protected Friend Overloads Function NewBitstreamException(ByVal errorcode As Integer) As BitstreamException
        Return New BitstreamException(errorcode, Nothing)
    End Function

    Protected Friend Overloads Function NewBitstreamException(ByVal errorcode As Integer, ByVal throwable As Exception) As BitstreamException
        Return New BitstreamException(errorcode, throwable)
    End Function

#End Region

End Class
