Imports System.IO
' Provides a view of the sequence of bytes that are produced during the conversion of an MP3 stream
' into a 16-bit PCM-encoded ("WAV" format) stream.
Public Class Mp3Stream
    Inherits Stream

#Region " Private Variables "
    Private m_Mp3Decoder As MP3Decoder
    Private m_BitStream As Bitstream
    Private m_SourceStream As Stream
    Private m_Frequency As Integer = -1
    Private m_OBuffer16BitStereo As OBuffer16BitStereo
#End Region

#Region " Properties "

    ' Gets the frequency of the audio being decoded.  
    ' Initially set to -1.  Initialized during the first call to either of the Read and DecodeFrames methods,
    ' and updated during every subsequent call to one of those methods to reflect the most recent header information
    ' from the MP3 stream.
    Public ReadOnly Property Frequency() As Integer
        Get
            Return m_Frequency
        End Get
    End Property

    ' Gets or sets the position of the source stream. This is relative to the number of bytes in the MP3 file, rather than
    ' the total number of PCM bytes (typically signicantly greater) contained in the Mp3Stream's output.
    Public Overrides Property Position() As Long
        Get
            Return m_SourceStream.Position
        End Get
        Set(ByVal Value As Long)
            m_SourceStream.Position = Value
        End Set
    End Property

#End Region

#Region " Constructors "

    ' Creates a new stream instance using the provided filename, and the default chunk size of 4096 bytes.
    Public Sub New(ByVal fileName As String)
        MyClass.New(New FileStream(fileName, FileMode.Open))
    End Sub

    ' Creates a new stream instance using the provided filename and chunk size.
    Public Sub New(ByVal fileName As String, ByVal chunkSize As Integer)
        MyClass.New(New FileStream(fileName, FileMode.Open), chunkSize)
    End Sub

    ' Creates a new stream instance using the provided stream as a source, and the default chunk size of 4096 bytes.
    Public Sub New(ByVal sourceStream As Stream)
        MyClass.New(sourceStream, 4096)
    End Sub

    ' Creates a new stream instance using the provided stream as a source.
    Public Sub New(ByVal sourceStream As Stream, ByVal chunkSize As Integer)
        m_Mp3Decoder = New MP3Decoder
        m_SourceStream = sourceStream
        m_BitStream = New Bitstream(New BackStream(m_SourceStream, chunkSize))
        m_OBuffer16BitStereo = New OBuffer16BitStereo

        m_Mp3Decoder.OutputBuffer = m_OBuffer16BitStereo
    End Sub

#End Region

#Region " Stream Implementation "

    Public Overrides ReadOnly Property CanRead() As Boolean
        Get
            Return m_SourceStream.CanRead
        End Get
    End Property

    Public Overrides ReadOnly Property CanSeek() As Boolean
        Get
            Return m_SourceStream.CanSeek
        End Get
    End Property

    Public Overrides ReadOnly Property CanWrite() As Boolean
        Get
            Return m_SourceStream.CanWrite
        End Get
    End Property

    Public Overrides ReadOnly Property Length() As Long
        Get
            Return m_SourceStream.Length
        End Get
    End Property

    Public Overrides Sub Flush()
        m_SourceStream.Flush()
    End Sub

#End Region

#Region " Functions "

    ' Sets the position of the source stream.
    Public Overrides Function Seek(ByVal pos As Long, ByVal origin As SeekOrigin) As Long
        Return m_SourceStream.Seek(pos, origin)
    End Function

    ' This method is not valid for an Mp3Stream.
    Public Overrides Sub SetLength(ByVal len As Long)
        Throw New InvalidOperationException
    End Sub

    ' This method is not valid for an Mp3Stream.
    Public Overrides Sub Write(ByVal buf() As Byte, ByVal ofs As Integer, ByVal count As Integer)
        Throw New InvalidOperationException
    End Sub

    ' Reads a frame from the MP3 stream.  Returns whether the operation was successful. If it wasn't, 
    ' the source stream is probably at its end.
    Private Function ReadFrame() As Boolean
        ' Read a frame from the bitstream.
        Dim m_FrameHeader As FrameHeader = m_BitStream.ReadFrame()
        If m_FrameHeader Is Nothing Then
            Return False
        End If
        Try
            ' Set the frequency value for the stream.
            m_Frequency = m_FrameHeader.Frequency()
            ' Decode the frame.
            Dim decoderOutput As AudioOutputBuffer = m_Mp3Decoder.DecodeFrame(m_FrameHeader, m_BitStream)
        Finally
            m_BitStream.CloseFrame()
        End Try
        Return True

    End Function

    ' Reads the MP3 stream as PCM-encoded bytes. Decodes a portion of the stream if necessary.
    ' Returns the number of bytes read.
    Public Overrides Function Read(ByVal buffer() As Byte, ByVal offset As Integer, ByVal count As Integer) As Integer
        ' Copy from queue buffers, reading new ones as necessary,
        ' until we can't read more or we have read "count" bytes
        Dim bytesRead As Integer
        While True
            If m_OBuffer16BitStereo.bytesLeft <= 0 Then
                If Not ReadFrame() Then
                    Exit While
                End If
            End If
            ' Copy as much as we can from the current buffer:
            bytesRead += m_OBuffer16BitStereo.Read(buffer, offset + bytesRead, count - bytesRead)

            If bytesRead >= count Then
                Exit While
            End If
        End While
        Return bytesRead
    End Function

#End Region

End Class



